{-# LANGUAGE TypeSynonymInstances #-}
 
module Value where

import Util

import Abstract

import Data.Set (Set)
import qualified Data.Set as Set
import Debug.Trace



-- call-by-value
-- cofuns are not forced

data Val = -- sizes
             VInfty
           | VSucc Val   
           | VMax [Val]
           | VMeta MVar Int          -- Xi + n  (n-fold successor of Xi)
           -- types   
           | VSort (Sort Val)
           | VPi Dec Name Val Env Expr
           | VSing Val TVal          -- Singleton type (TVal is base type)
           -- functions
           | VLam Name Env Expr
           | VUp Val TVal            -- delayed eta expansion; TVal is a VPi
           -- neutrals
           | VGen Int                -- free variable (de Bruijn level)
           | VDef DefId              -- co(data/constructor/fun)
                                     -- VDef occurs only inside a VApp!

{-
           | VCon Co Name
           | VDef Name               -- data/codata, fun/cofun identifier
-}
           | VCase Val Env [Clause]
           | VApp Val [Clos]
           | VClos Env Expr          -- closure for cbn evaluation
           -- don't care
           | VIrr                    -- erased hypothetical inhabitant of empty type
             deriving (Eq,Ord)

type TVal = Val
type Clos = Val

-- smart constructors ------------------------------------------------

mkClos :: Env -> Expr -> Val
mkClos rho (Var x) = lookupPure rho x 
mkClos rho e = VClos (filterEnv (freeVars e) rho) e

filterEnv :: Set Name -> Env -> Env
filterEnv ns [] = []
filterEnv ns ((x,v) : rho) = 
  if Set.member x ns then (x,v) : filterEnv (Set.delete x ns) rho 
   else filterEnv ns rho  

vDef id   = VDef id `VApp` []
vCon co n = vDef $ DefId (Con co) n
vFun n    = vDef $ DefId Fun n
vDat n    = vDef $ DefId Dat n

failValInv :: (Monad m) => Val -> m a
failValInv v = fail $ "internal error: value " ++ show v ++ " violates representation invariant"

arrow :: TVal -> TVal -> TVal
arrow a b = VPi defaultDec "" a [(bla,b)] (Var bla)
  where bla = "#ARROW#"

-- a wellformed size value takes one of the following three forms
-- * VInfty
-- * VSucc^* VGen
-- * VMax vs where each v_i = VSucc^* (VGen k_i) and all k_i different
--           and vs has length >= 2
--
-- the smart constructors construct wellformed size values using the laws
-- $ #             = #                Infty
-- max # k         = #                
-- $ (max i j)     = max ($ i) ($ j)  $ distributes over max
-- max (max i j) k = max i j k        Assoc-Commut of max
-- max i i         = i                Idempotency of max
succSize :: Val -> Val
succSize v = case v of
            VInfty -> VInfty
            VMax vs -> VMax $ map succSize vs
            VMeta i n -> VMeta i (n + 1)
            _ -> VSucc v               

-- maxSize vs = VInfty                 if any v_i=Infty
--            = VMax (sort (nub (flatten vs)) else
-- precondition vs 
maxSize :: [Val] -> Val
maxSize vs = case flatten [] vs of
   [] -> VInfty
   [v] -> v
   vs' -> VMax vs' 
  where flatten acc (VInfty:_) = []
        flatten acc (VMax vs:vs') = flatten (vs ++ acc) vs'
        flatten acc (v:vs) = flatten (v:acc) vs
        flatten acc [] = Set.toList $ Set.fromList acc -- sort, nub

-- destructors -------------------------------------------------------

vSortToSort :: Sort Val -> Sort Expr
vSortToSort Type = Type
vSortToSort Size = Size
vSortToSort Tm   = Tm
vSortToSort Set  = Set

predSize :: Val -> Maybe Val
predSize VInfty = Just VInfty
predSize (VSucc v) = Just v
predSize (VMax vs) = do vs' <- mapM predSize vs
                        return $ maxSize vs'
predSize _ = Nothing

isFunType :: TVal -> Bool
isFunType (VPi{}) = True
isFunType _ = False

isDataType :: TVal -> Bool
isDataType (VApp (VDef (DefId Dat _)) _) = True
isDataType (VSing v tv) = isDataType tv
isDataType _ = False

-- ugly printing -----------------------------------------------------

instance Show Val where
    show = showVal

showVal :: Val -> String
showVal (VSort s) = show s
showVal VInfty = "#"
showVal (VSucc v) = "($ " ++ showVal v ++ ")" 
showVal (VMax vl) = "(max " ++ showVals vl ++ ")"
showVal (VApp v []) = showVal v
showVal (VApp v vl) = "(" ++ showVal v ++ " " ++ showVals vl ++ ")"
-- showVal (VCon _ n) = n
showVal (VDef id) = name id
showVal (VGen k) = "v" ++ show k
showVal (VMeta k 0) = "?" ++ show k
showVal (VMeta k 1) = "$?" ++ show k
showVal (VMeta k n) = "(?" ++ show k ++ "+" ++ show n ++")"
showVal (VCase v env cs) = "case " ++ showVal v ++ "{ " ++ showCases cs ++ " } " ++ showEnv env 
showVal (VLam x env e) = "(\\" ++ x ++ " -> " ++ show e ++ showEnv env ++ ")" 
showVal (VClos [] e) = showsPrec precAppR e ""
showVal (VClos env e) = "{" ++ show e ++ " " ++ showEnv env ++ "}" 
showVal (VUp v vt) = "(" ++ show v ++ " Up " ++ show vt ++ ")" 
showVal (VSing v vt) = "<" ++ show v ++ " : " ++ show vt ++ ">" 
showVal VIrr  = "*"

showVal (VPi dec x av env b) =
  parens $ (if erased dec then brackets binding
              else if x=="" then s1 else parens binding) 
                                  ++ " -> " ++ show b ++ showEnv env
         where s1 = showVal av
               binding = if x == "" then s1 else x ++ " : " ++ s1 

{-
showVal (VPi dec "" av env e) = "(" ++ showVal av ++ " -> " ++ show e ++ showEnv env ++ ")"
showVal (VPi dec x av env e) = "((" ++ x  ++ " : " ++ showVal av ++ ") -> " ++ show e ++ showEnv env ++ ")"
-}

showVals :: [Val] -> String
showVals [] = ""
showVals (v:vl) = showVal v ++ (if null vl then "" else " " ++ showVals vl)   

-- environment ---------------------------------------------------

type Environ a = [(Name,a)]
type Env = Environ Val

emptyEnv = []

-- enviroment extension / update
update :: Environ a -> Name -> a -> Environ a
update env "" v = env
update env n v = (n,v):env

lookupPure :: Show a => Environ a -> Name -> a
lookupPure rho x = 
    case lookup x rho of
      Just v -> v
      Nothing -> error $ "lookupPure: unbound identifier " ++ x ++ " in environment " ++ show rho

lookupEnv :: Monad m => Environ a -> Name -> m a
lookupEnv [] n = fail $ "lookupEnv: identifier " ++ show n ++ " not bound" 
lookupEnv ((x,v):xs) n = if x == n then return v 
                          else lookupEnv xs n

showEnv :: Environ Val -> String
showEnv [] = ""
showEnv x = "{" ++ showEnv' x ++ "}" where
    showEnv' ((n,v):env) = n ++ " = " ++ showVal v ++ 
      (if null env then "" else ", " ++ showEnv' env) 

fresh :: Environ a -> Name
fresh env = "fresh#" ++ show (length env)
