-- MiniAgda encoding of (most of) Negative.agda. Uses code written by
-- Andreas Abel. /NAD

-- version with standard sizing of Eq

data Nat : Set
{ zero : Nat
; suc  : Nat -> Nat
}

data Id (A : Set) (x : A) : A -> Set
{ refl : Id A x x
}

data Type : Set
{ nat : Type
; arr : Type -> Type -> Type
}

data Ctxt : Set
{ empty : Ctxt
; snoc  : Ctxt -> Type -> Ctxt
}

data Tm : Ctxt -> Type -> Set
{ vz  : [G : Ctxt] ->               [b : Type] -> Tm (snoc G b) b
; wk  : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm G b -> Tm (snoc G a) b
; lam : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm (snoc G a) b -> Tm G (arr a b)
; app : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm G (arr a b) -> Tm G a -> Tm G b
; fix : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm G (arr (arr a b) (arr a b)) -> Tm G a -> Tm G b
}

data Flag : Set
{ env : Flag
; val : Flag
}

fun Index : Flag -> Set
{ Index env = Ctxt
; Index val = Type
}

data EnvOrVal : (f : Flag) -> Index f -> Set
{ emptyE : EnvOrVal env empty
; snocE  : [G : Ctxt] -> [b : Type] -> EnvOrVal env G -> EnvOrVal val b -> EnvOrVal env (snoc G b)
; natV   : Nat -> EnvOrVal val nat
; arrV   : [G : Ctxt] -> [a : Type] -> [b : Type] -> Tm (snoc G a) b -> EnvOrVal env G -> EnvOrVal val (arr a b)
}

fun Env : Ctxt -> Set
{ Env = EnvOrVal env
}

fun Value : Type -> Set
{ Value = EnvOrVal val
}

sized codata Lift (+ A : Set) : Size -> Set
{ now   : [i : Size] -> A        -> Lift A ($ i)
; later : [i : Size] -> Lift A i -> Lift A ($ i)
}

fun bind : [A : Set] -> [B : Set] ->
           [i : Size] -> Lift A i -> (A -> Lift B i) -> Lift B i
{ bind A B ($ i) (now   .A .i a) f = f a
; bind A B ($ i) (later .A .i x) f = later B i (bind A B i x f)
}

-- I have defined strong bisimilarity instead of weak here.

sized codata Eq (A : Set) (+ eq : A -> A -> Set) :
                (i : Size) -> Lift A # -> Lift A # -> Set
{ nowCong   : [i : Size] -> (v1 : A) -> (v2 : A) -> eq v1 v2 ->
              Eq A eq ($ i) (now A # v1) (now A # v2)
; laterCong : [i : Size] -> [x1 : Lift A #] -> [x2 : Lift A #] ->
              Eq A eq i x1 x2 -> Eq A eq ($ i) (later A # x1) (later A # x2)
}

fun bindCong : [i : Size] -> [A : Set] -> [B : Set] ->
               [eqA : A -> A -> Set] -> [eqB : B -> B -> Set] ->
               [x1 : Lift A #] -> [x2 : Lift A #] ->
               [f1 : A -> Lift B #] -> [f2 : A -> Lift B #] ->
               Eq A eqA i x1 x2 ->
               ((v1 : A) -> (v2 : A) -> eqA v1 v2 -> Eq B eqB i (f1 v1) (f2 v2)) ->
               Eq B eqB i (bind A B # x1 f1) (bind A B # x2 f2)
{ bindCong ($ i) A B EqA EqB .(now A # v1) .(now A # v2) f1 f2
   (nowCong .A .EqA .i v1 v2 v1v2) f1f2 = f1f2 v1 v2 v1v2
; bindCong ($ i) A B EqA EqB .(later A # x1) .(later A # x2) f1 f2
   (laterCong .A .EqA .i x1 x2 x1x2) f1f2 =
   laterCong B EqB i (bind A B # x1 f1) (bind A B # x2 f2)
     (bindCong i A B EqA EqB x1 x2 f1 f2 x1x2 f1f2)
}

mutual {

  fun den : [i : Size] -> [G : Ctxt] -> [b : Type] -> Tm G b -> Env G -> Lift (Value b) i
  { den i     .(snoc G b) .b         (vz G b)          (snocE .G .b rho v) = now (Value b) i v
  ; den i     .(snoc G a) .b         (wk G a b t)      (snocE .G .a rho v) = den i G b t rho
  ; den i     .G          .(arr a b) (lam G a b t)     rho                 = now (Value (arr a b)) i (arrV G a b t rho)
  ; den i     .G          .b         (app G a b t1 t2) rho =
     bind (Value (arr a b)) (Value b) i (den i G (arr a b) t1 rho) (\ f ->
     bind (Value a)         (Value b) i (den i G a         t2 rho) (\ v ->
     appV i a b f v))
  ; den ($ i) .G          .b         (fix G a b t1 t2) rho =
     later (Value b) i
       (den i G b (app G a b
                     (app G (arr a b) (arr a b)
                        t1
                        (lam G a b
                           (fix (snoc G a) a b
                              (wk G a (arr (arr a b) (arr a b)) t1)
                              (vz G a))))
                     t2) rho)
  }

  fun appV : [i : Size] -> [a : Type] -> [b : Type] -> Value (arr a b) -> Value a -> Lift (Value b) i
  { appV ($ i) .a .b (arrV G a b t1 rho) v2 =
     later (Value b) i (den i (snoc G a) b t1 (snocE G a rho v2))
  }

}

mutual {

  fun EqL : (- i : Size) -> (a : Type) -> Lift (Value a) # -> Lift (Value a) # -> Set
  { EqL i a x y = Eq (Value a) (EqV i a) i x y
  }

  -- Note: In Negative.agda I used \infty in the last clause of
  -- _\approx{}V_. This is not needed for this development.

  fun EqV : (- i : Size) -> (a : Type) -> Value a -> Value a -> Set
  { EqV i nat       (natV x)                (natV y)                = Id Nat x y
  ; EqV i (arr a b) (arrV G1 .a .b t1 rho1) (arrV G2 .a .b t2 rho2) =
      (v1 : Value a) -> (v2 : Value a) ->
      EqV # a v1 v2 ->
      EqL i b (den # (snoc G1 a) b t1 (snocE G1 a rho1 v1))
              (den # (snoc G2 a) b t2 (snocE G2 a rho2 v2))
  }

}

data EqEnv : (G : Ctxt) -> Env G -> Env G -> Set
{ emptyEq : EqEnv empty emptyE emptyE
; snocEq  : [G : Ctxt] -> [b : Type] ->
            [rho1 : Env G] -> [rho2 : Env G] ->
            [v1 : Value b] -> [v2 : Value b] ->
            EqEnv G rho1 rho2 -> EqV # b v1 v2 ->
            EqEnv (snoc G b) (snocE G b rho1 v1) (snocE G b rho2 v2)
}

fun EqT : (G : Ctxt) -> (b : Type) -> Tm G b -> Tm G b -> Set
{ EqT G b t1 t2 = [i : Size] -> (rho1 : Env G) -> (rho2 : Env G) ->
                  EqEnv G rho1 rho2 ->
                  EqL i b (den # G b t1 rho1) (den # G b t2 rho2)
}

fun appCong : [i : Size] -> [a : Type] -> [b : Type] ->
              [f1 : Value (arr a b)] -> [f2 : Value (arr a b)] ->
              (v1 : Value a) -> (v2 : Value a) ->
              EqV i (arr a b) f1 f2 -> EqV # a v1 v2 ->
              EqL i b (appV # a b f1 v1) (appV # a b f2 v2)
{ appCong ($ i) a b (arrV G1 .a .b t1 rho1) (arrV G2 .a .b t2 rho2) v1 v2 f1f2 v1v2 =
    laterCong (Value b) (EqV ($ i) b) i
              (den # (snoc G1 a) b t1 (snocE G1 a rho1 v1))
              (den # (snoc G2 a) b t2 (snocE G2 a rho2 v2))
              (f1f2 v1 v2 v1v2)
}

fun reflT : [G : Ctxt] -> [b : Type] -> (t : Tm G b) -> EqT G b t t
{ reflT .(snoc G b) .b (vz G b) i
        (snocE .G .b rho1 v1)
        (snocE .G .b rho2 v2)
        (snocEq .G .b .rho1 .rho2 .v1 .v2 rhoEq vEq) =
    nowCong (Value b) (EqV # b) i v1 v2 vEq
; reflT .(snoc G a) .b (wk G a b t) i
        (snocE .G .a rho1 v1)
        (snocE .G .a rho2 v2)
        (snocEq .G .a .rho1 .rho2 .v1 .v2 rhoEq vEq) =
    reflT G b t i rho1 rho2 rhoEq
; reflT .G .(arr a b) (lam G a b t) i rho1 rho2 rhoEq =
    nowCong (Value (arr a b)) (EqV i (arr a b)) i
            (arrV G a b t rho1)
            (arrV G a b t rho2)
      (\ v1 -> \ v2 -> \ vEq ->
         reflT (snoc G a) b t i
               (snocE G a rho1 v1)
               (snocE G a rho2 v2)
               (snocEq G a rho1 rho2 v1 v2 rhoEq vEq))
; reflT .G .b (app G a b t1 t2) i rho1 rho2 rhoEq =
   bindCong i (Value (arr a b)) (Value b) (EqV # (arr a b)) (EqV i b)
            (den # G (arr a b) t1 rho1)
            (den # G (arr a b) t1 rho2)
            (\ f -> bind (Value a) (Value b) # (den # G a t2 rho1) (\ v -> appV # a b f v))
            (\ f -> bind (Value a) (Value b) # (den # G a t2 rho2) (\ v -> appV # a b f v))
            (reflT G (arr a b) t1 i rho1 rho2 rhoEq)
            (\ f1 -> \ f2 -> \ fEq ->
   bindCong i (Value a) (Value b) (EqV # a) (EqV i b)
            (den # G a t2 rho1)
            (den # G a t2 rho2)
            (appV # a b f1)
            (appV # a b f2)
            (reflT G a t2 i rho1 rho2 rhoEq)
            (\ v1 -> \ v2 -> \ vEq ->
   appCong i a b f1 f2 v1 v2 fEq vEq))
; reflT .G .b (fix G a b t1 t2) ($ i) rho1 rho2 rhoEq =
    let t : Tm G b = app G a b
                       (app G (arr a b) (arr a b)
                          t1
                          (lam G a b
                             (fix (snoc G a) a b
                                (wk G a (arr (arr a b) (arr a b)) t1)
                                (vz G a))))
                       t2 in
    laterCong (Value b) (EqV i b) i
              (den # G b t rho1)
              (den # G b t rho2)
              (reflT G b t i rho1 rho2 rhoEq)
}
